;; Generated package description from javaimp.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "javaimp" "0.9.2.0.20260609.5" "Add and reorder Java import statements in Maven/Gradle projects" 'nil :commit "b52e1cdba17e1b1d31750cb123ef724788fbd0d5" :url "https://elpa.gnu.org/packages/javaimp.html" :authors '(("Filipp Gunbin" . "fgunbin@fastmail.fm")) :maintainer '("Filipp Gunbin" . "fgunbin@fastmail.fm") :keywords '("java" "maven" "gradle" "programming"))
